# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from datetime import datetime
from typing import Any, Dict, Optional

from truefoundry.ml._autogen.client.models.artifact_version_serialization_format import (
    ArtifactVersionSerializationFormat,
)
from truefoundry.ml._autogen.client.models.artifact_version_status import (
    ArtifactVersionStatus,
)
from truefoundry.ml._autogen.client.models.manifest1 import Manifest1
from truefoundry.ml._autogen.client.models.subject import Subject
from truefoundry.pydantic_v1 import BaseModel, Field, StrictInt, StrictStr


class ArtifactVersionDto(BaseModel):
    """
    ArtifactVersionDto
    """

    id: StrictStr = Field(...)
    artifact_id: StrictStr = Field(...)
    version: StrictInt = Field(...)
    fqn: StrictStr = Field(...)
    experiment_id: StrictStr = Field(...)
    artifact_storage_root: Optional[StrictStr] = None
    artifact_metadata: Optional[Dict[str, Any]] = None
    description: Optional[StrictStr] = None
    status: Optional[ArtifactVersionStatus] = None
    step: Optional[StrictInt] = 0
    created_by: Optional[StrictStr] = None
    created_by_subject: Subject = Field(...)
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    artifact_fqn: Optional[StrictStr] = None
    data_path: Optional[StrictStr] = None
    usage_code_snippet: Optional[StrictStr] = None
    run_id: Optional[StrictStr] = None
    artifact_name: Optional[StrictStr] = None
    internal_metadata: Optional[Dict[str, Any]] = None
    artifact_size: Optional[StrictInt] = None
    manifest: Optional[Manifest1] = None
    serialization_format: ArtifactVersionSerializationFormat = Field(...)
    __properties = [
        "id",
        "artifact_id",
        "version",
        "fqn",
        "experiment_id",
        "artifact_storage_root",
        "artifact_metadata",
        "description",
        "status",
        "step",
        "created_by",
        "created_by_subject",
        "created_at",
        "updated_at",
        "artifact_fqn",
        "data_path",
        "usage_code_snippet",
        "run_id",
        "artifact_name",
        "internal_metadata",
        "artifact_size",
        "manifest",
        "serialization_format",
    ]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ArtifactVersionDto:
        """Create an instance of ArtifactVersionDto from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of created_by_subject
        if self.created_by_subject:
            _dict["created_by_subject"] = self.created_by_subject.to_dict()
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of manifest
        if self.manifest:
            _dict["manifest"] = self.manifest.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> ArtifactVersionDto:
        """Create an instance of ArtifactVersionDto from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return ArtifactVersionDto.parse_obj(obj)

        _obj = ArtifactVersionDto.parse_obj(
            {
                "id": obj.get("id"),
                "artifact_id": obj.get("artifact_id"),
                "version": obj.get("version"),
                "fqn": obj.get("fqn"),
                "experiment_id": obj.get("experiment_id"),
                "artifact_storage_root": obj.get("artifact_storage_root"),
                "artifact_metadata": obj.get("artifact_metadata"),
                "description": obj.get("description"),
                "status": obj.get("status"),
                "step": obj.get("step") if obj.get("step") is not None else 0,
                "created_by": obj.get("created_by"),
                "created_by_subject": Subject.from_dict(obj.get("created_by_subject"))
                if obj.get("created_by_subject") is not None
                else None,
                "created_at": obj.get("created_at"),
                "updated_at": obj.get("updated_at"),
                "artifact_fqn": obj.get("artifact_fqn"),
                "data_path": obj.get("data_path"),
                "usage_code_snippet": obj.get("usage_code_snippet"),
                "run_id": obj.get("run_id"),
                "artifact_name": obj.get("artifact_name"),
                "internal_metadata": obj.get("internal_metadata"),
                "artifact_size": obj.get("artifact_size"),
                "manifest": Manifest1.from_dict(obj.get("manifest"))
                if obj.get("manifest") is not None
                else None,
                "serialization_format": obj.get("serialization_format"),
            }
        )
        return _obj



from typing import Any, List, Optional

from truefoundry.deploy.io.output_callback import OutputCallBack


class NoOutputCallBack(OutputCallBack):
    def print_header(self, line: Any) -> None:
        pass

    def _print_separator(self) -> None:
        pass

    def print_line(self, line: str) -> None:
        pass

    def print_lines_in_panel(
        self, lines: List[str], header: Optional[str] = None
    ) -> None:
        pass

    def print_code_in_panel(
        self, lines: List[str], header: Optional[str] = None
    ) -> None:
        pass

    def print(self, line: Any) -> None:
        pass



INTERNAL_ERROR = 1
INVALID_PARAMETER_VALUE = 1000


class MlflowException(Exception):
    """
    Generic exception thrown to surface failure information about external-facing operations.
    The error message associated with this exception may be exposed to clients in HTTP responses
    for debugging purposes. If the error text is sensitive, raise a generic `Exception` object
    instead.
    """

    def __init__(self, message, error_code=INTERNAL_ERROR, **kwargs):
        """
        :param message: The message describing the error that occured. This will be included in the
                        exception's serialized JSON representation.
        :param error_code: An appropriate error code for the error that occured; it will be included
                           in the exception's serialized JSON representation. This should be one of
                           the codes listed in the `mlflow.protos.databricks_pb2` proto.
        :param kwargs: Additional key-value pairs to include in the serialized JSON representation
                       of the MlflowException.
        """
        self.error_code = error_code
        self.message = str(message)
        self.json_kwargs = kwargs
        super().__init__(message)

    @classmethod
    def invalid_parameter_value(cls, message, **kwargs):
        return cls(message, error_code=INVALID_PARAMETER_VALUE, **kwargs)



import rich_click as click

from truefoundry.cli.const import GROUP_CLS
from truefoundry.ml.cli.commands.model_init import get_model_init_command


@click.group(
    name="deploy-init",
    cls=GROUP_CLS,
    help="Initialize the TrueFoundry deployment configuration.",
)
def deploy_init_command(): ...


def get_deploy_init_command():
    """
    Generates the deploy-init command.
        model: Initialize the TrueFoundry deployment configuration for a model. eg: tfy deploy-init model [--args]
    """

    deploy_init_command.add_command(get_model_init_command())
    return deploy_init_command



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING, Any, List, Optional, Union

from truefoundry.ml._autogen.client.models.fast_ai_framework import FastAIFramework
from truefoundry.ml._autogen.client.models.gluon_framework import GluonFramework
from truefoundry.ml._autogen.client.models.h2_o_framework import H2OFramework
from truefoundry.ml._autogen.client.models.keras_framework import KerasFramework
from truefoundry.ml._autogen.client.models.light_gbm_framework import LightGBMFramework
from truefoundry.ml._autogen.client.models.onnx_framework import ONNXFramework
from truefoundry.ml._autogen.client.models.paddle_framework import PaddleFramework
from truefoundry.ml._autogen.client.models.py_torch_framework import PyTorchFramework
from truefoundry.ml._autogen.client.models.sklearn_framework import SklearnFramework
from truefoundry.ml._autogen.client.models.spa_cy_framework import SpaCyFramework
from truefoundry.ml._autogen.client.models.stats_models_framework import (
    StatsModelsFramework,
)
from truefoundry.ml._autogen.client.models.tensor_flow_framework import (
    TensorFlowFramework,
)
from truefoundry.ml._autogen.client.models.transformers_framework import (
    TransformersFramework,
)
from truefoundry.ml._autogen.client.models.xg_boost_framework import XGBoostFramework
from truefoundry.pydantic_v1 import (
    BaseModel,
    Field,
    ValidationError,
    validator,
)

FRAMEWORK_ANY_OF_SCHEMAS = [
    "FastAIFramework",
    "GluonFramework",
    "H2OFramework",
    "KerasFramework",
    "LightGBMFramework",
    "ONNXFramework",
    "PaddleFramework",
    "PyTorchFramework",
    "SklearnFramework",
    "SpaCyFramework",
    "StatsModelsFramework",
    "TensorFlowFramework",
    "TransformersFramework",
    "XGBoostFramework",
]


class Framework(BaseModel):
    """
    Framework for the model version like Transformers, PyTorch, Sklearn, Xgboost etc with framework specific metadata. This will be used to infer model deployment configuration
    """

    # data type: TransformersFramework
    anyof_schema_1_validator: Optional[TransformersFramework] = None
    # data type: TensorFlowFramework
    anyof_schema_2_validator: Optional[TensorFlowFramework] = None
    # data type: SklearnFramework
    anyof_schema_3_validator: Optional[SklearnFramework] = None
    # data type: PyTorchFramework
    anyof_schema_4_validator: Optional[PyTorchFramework] = None
    # data type: KerasFramework
    anyof_schema_5_validator: Optional[KerasFramework] = None
    # data type: XGBoostFramework
    anyof_schema_6_validator: Optional[XGBoostFramework] = None
    # data type: LightGBMFramework
    anyof_schema_7_validator: Optional[LightGBMFramework] = None
    # data type: FastAIFramework
    anyof_schema_8_validator: Optional[FastAIFramework] = None
    # data type: H2OFramework
    anyof_schema_9_validator: Optional[H2OFramework] = None
    # data type: ONNXFramework
    anyof_schema_10_validator: Optional[ONNXFramework] = None
    # data type: SpaCyFramework
    anyof_schema_11_validator: Optional[SpaCyFramework] = None
    # data type: StatsModelsFramework
    anyof_schema_12_validator: Optional[StatsModelsFramework] = None
    # data type: GluonFramework
    anyof_schema_13_validator: Optional[GluonFramework] = None
    # data type: PaddleFramework
    anyof_schema_14_validator: Optional[PaddleFramework] = None
    if TYPE_CHECKING:
        actual_instance: Union[
            FastAIFramework,
            GluonFramework,
            H2OFramework,
            KerasFramework,
            LightGBMFramework,
            ONNXFramework,
            PaddleFramework,
            PyTorchFramework,
            SklearnFramework,
            SpaCyFramework,
            StatsModelsFramework,
            TensorFlowFramework,
            TransformersFramework,
            XGBoostFramework,
        ]
    else:
        actual_instance: Any
    any_of_schemas: List[str] = Field(FRAMEWORK_ANY_OF_SCHEMAS, const=True)

    class Config:
        validate_assignment = True

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                raise ValueError(
                    "If a position argument is used, only 1 is allowed to set `actual_instance`"
                )
            if kwargs:
                raise ValueError(
                    "If a position argument is used, keyword arguments cannot be used."
                )
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(**kwargs)

    @validator("actual_instance")
    def actual_instance_must_validate_anyof(cls, v):
        instance = Framework.construct()
        error_messages = []
        # validate data type: TransformersFramework
        if not isinstance(v, TransformersFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `TransformersFramework`"
            )
        else:
            return v

        # validate data type: TensorFlowFramework
        if not isinstance(v, TensorFlowFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `TensorFlowFramework`"
            )
        else:
            return v

        # validate data type: SklearnFramework
        if not isinstance(v, SklearnFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `SklearnFramework`"
            )
        else:
            return v

        # validate data type: PyTorchFramework
        if not isinstance(v, PyTorchFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `PyTorchFramework`"
            )
        else:
            return v

        # validate data type: KerasFramework
        if not isinstance(v, KerasFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `KerasFramework`"
            )
        else:
            return v

        # validate data type: XGBoostFramework
        if not isinstance(v, XGBoostFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `XGBoostFramework`"
            )
        else:
            return v

        # validate data type: LightGBMFramework
        if not isinstance(v, LightGBMFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `LightGBMFramework`"
            )
        else:
            return v

        # validate data type: FastAIFramework
        if not isinstance(v, FastAIFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `FastAIFramework`"
            )
        else:
            return v

        # validate data type: H2OFramework
        if not isinstance(v, H2OFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `H2OFramework`"
            )
        else:
            return v

        # validate data type: ONNXFramework
        if not isinstance(v, ONNXFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `ONNXFramework`"
            )
        else:
            return v

        # validate data type: SpaCyFramework
        if not isinstance(v, SpaCyFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `SpaCyFramework`"
            )
        else:
            return v

        # validate data type: StatsModelsFramework
        if not isinstance(v, StatsModelsFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `StatsModelsFramework`"
            )
        else:
            return v

        # validate data type: GluonFramework
        if not isinstance(v, GluonFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `GluonFramework`"
            )
        else:
            return v

        # validate data type: PaddleFramework
        if not isinstance(v, PaddleFramework):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `PaddleFramework`"
            )
        else:
            return v

        if error_messages:
            # no match
            raise ValueError(
                "No match found when setting the actual_instance in Framework with anyOf schemas: FastAIFramework, GluonFramework, H2OFramework, KerasFramework, LightGBMFramework, ONNXFramework, PaddleFramework, PyTorchFramework, SklearnFramework, SpaCyFramework, StatsModelsFramework, TensorFlowFramework, TransformersFramework, XGBoostFramework. Details: "
                + ", ".join(error_messages)
            )
        else:
            return v

    @classmethod
    def from_dict(cls, obj: dict) -> Framework:
        return cls.from_json(json.dumps(obj))

    @classmethod
    def from_json(cls, json_str: str) -> Framework:
        """Returns the object represented by the json string"""
        instance = Framework.construct()
        error_messages = []
        # anyof_schema_1_validator: Optional[TransformersFramework] = None
        try:
            instance.actual_instance = TransformersFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_2_validator: Optional[TensorFlowFramework] = None
        try:
            instance.actual_instance = TensorFlowFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_3_validator: Optional[SklearnFramework] = None
        try:
            instance.actual_instance = SklearnFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_4_validator: Optional[PyTorchFramework] = None
        try:
            instance.actual_instance = PyTorchFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_5_validator: Optional[KerasFramework] = None
        try:
            instance.actual_instance = KerasFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_6_validator: Optional[XGBoostFramework] = None
        try:
            instance.actual_instance = XGBoostFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_7_validator: Optional[LightGBMFramework] = None
        try:
            instance.actual_instance = LightGBMFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_8_validator: Optional[FastAIFramework] = None
        try:
            instance.actual_instance = FastAIFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_9_validator: Optional[H2OFramework] = None
        try:
            instance.actual_instance = H2OFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_10_validator: Optional[ONNXFramework] = None
        try:
            instance.actual_instance = ONNXFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_11_validator: Optional[SpaCyFramework] = None
        try:
            instance.actual_instance = SpaCyFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_12_validator: Optional[StatsModelsFramework] = None
        try:
            instance.actual_instance = StatsModelsFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_13_validator: Optional[GluonFramework] = None
        try:
            instance.actual_instance = GluonFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_14_validator: Optional[PaddleFramework] = None
        try:
            instance.actual_instance = PaddleFramework.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))

        if error_messages:
            # no match
            raise ValueError(
                "No match found when deserializing the JSON string into Framework with anyOf schemas: FastAIFramework, GluonFramework, H2OFramework, KerasFramework, LightGBMFramework, ONNXFramework, PaddleFramework, PyTorchFramework, SklearnFramework, SpaCyFramework, StatsModelsFramework, TensorFlowFramework, TransformersFramework, XGBoostFramework. Details: "
                + ", ".join(error_messages)
            )
        else:
            return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        to_json = getattr(self.actual_instance, "to_json", None)
        if callable(to_json):
            return self.actual_instance.to_json()
        else:
            return json.dumps(self.actual_instance)

    def to_dict(self) -> dict:
        """Returns the dict representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        to_json = getattr(self.actual_instance, "to_json", None)
        if callable(to_json):
            return self.actual_instance.to_dict()
        else:
            # primitive type
            return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance"""
        return pprint.pformat(self.dict())



import os
import re
import sys
from typing import Dict, Optional

import click
import docker
import questionary
import requests
import yaml
from dotenv import dotenv_values
from openai import OpenAI
from rich.console import Console
from rich.prompt import Prompt
from rich.status import Status

from truefoundry.autodeploy.agents.developer import Developer
from truefoundry.autodeploy.agents.project_identifier import (
    ComponentType,
    ProjectIdentifier,
)
from truefoundry.autodeploy.agents.tester import Tester
from truefoundry.autodeploy.constants import (
    ABOUT_AUTODEPLOY,
    AUTODEPLOY_OPENAI_API_KEY,
    AUTODEPLOY_OPENAI_BASE_URL,
    AUTODEPLOY_TFY_BASE_URL,
)
from truefoundry.autodeploy.exception import InvalidRequirementsException
from truefoundry.autodeploy.tools.ask import AskQuestion
from truefoundry.autodeploy.tools.commit import CommitConfirmation
from truefoundry.autodeploy.tools.docker_run import DockerRun, DockerRunLog
from truefoundry.autodeploy.utils.client import get_git_binary
from truefoundry.cli.const import COMMAND_CLS
from truefoundry.common.session import Session
from truefoundry.deploy import Build, DockerFileBuild, Job, LocalSource, Port, Service
from truefoundry.deploy.lib.clients.servicefoundry_client import (
    ServiceFoundryServiceClient,
)


def _get_openai_client() -> OpenAI:
    if AUTODEPLOY_OPENAI_BASE_URL is not None and AUTODEPLOY_OPENAI_API_KEY is not None:
        return OpenAI(
            api_key=AUTODEPLOY_OPENAI_API_KEY, base_url=AUTODEPLOY_OPENAI_BASE_URL
        )
    try:
        session = Session.new()
        resp = requests.get(
            f"{AUTODEPLOY_TFY_BASE_URL}/api/svc/v1/llm-gateway/access-details",
            headers={
                "Authorization": f"Bearer {session.access_token}",
            },
        )
        resp.raise_for_status()
        resp = resp.json()
        return OpenAI(api_key=resp["jwtToken"], base_url=resp["inferenceBaseURL"])
    except requests.exceptions.HTTPError as http_error:
        if http_error.response.status_code in [401, 403]:
            raise InvalidRequirementsException(
                message="Unauthorized access to TrueFoundry server. "
                "Please verify your credentials and ensure you have the necessary access rights."
                "\nIf you wish to proceed without TrueFoundry AI, "
                "you need to either have a truefoundry.yaml file in your project root "
                'or pass the path to a yaml file using the "--file file_name" option.'
            ) from http_error

        raise http_error

    except Exception as e:
        raise InvalidRequirementsException(message=str(e)) from e


def _get_service_port(name: str, workspace_fqn: str, port: int) -> Port:
    client = ServiceFoundryServiceClient()
    workspaces = client.get_workspace_by_fqn(workspace_fqn=workspace_fqn)
    if len(workspaces) == 0:
        raise click.UsageError("Invalid workspace")
    workspace = workspaces[0]
    cluster = client.get_cluster(cluster_id=workspace.clusterId)

    generated_name = name + "-" + workspace.name
    base_domain_urls = cluster["metadata"].get("baseDomainURLs")
    if not base_domain_urls:
        return Port(port=port, expose=False)

    base_domain_url = base_domain_urls[0]

    if "*" in base_domain_url:
        return Port(
            port=port, host=base_domain_url.replace("*", generated_name).strip("/")
        )

    return Port(port=port, host=base_domain_url, path=f"/{generated_name}/")


def deploy_component(
    workspace_fqn: str,
    project_root_path: str,
    dockerfile_path: str,
    component_type: ComponentType,
    name: str,
    env: Dict,
    command: Optional[str] = None,
    port: Optional[int] = None,
):
    if not os.path.exists(os.path.join(project_root_path, dockerfile_path)):
        raise FileNotFoundError("Dockerfile not found in the project.")

    image = Build(
        build_spec=DockerFileBuild(
            dockerfile_path=dockerfile_path,
            command=command,
        ),
        build_source=LocalSource(project_root_path=project_root_path),
    )
    if component_type == ComponentType.SERVICE:
        if port is None:
            raise ValueError("Port is required for deploying service")
        service_port = _get_service_port(
            name=name, workspace_fqn=workspace_fqn, port=port
        )

        app = Service(
            name=name,
            image=image,
            ports=[service_port],
            env=env,
        )
    else:
        app = Job(name=name, image=image, env=env)
    with open("truefoundry.yaml", "w") as application_spec:
        yaml.dump(app.dict(), application_spec, indent=2)
    app.deploy(workspace_fqn=workspace_fqn)


def _parse_env(project_root_path: str, env_path: str) -> Dict:
    if not os.path.isabs(env_path):
        env_path = os.path.join(project_root_path, env_path)

    if os.path.exists(env_path):
        return dotenv_values(env_path)

    raise FileNotFoundError(f"Invalid path {env_path!r}")


def _check_repo(project_root_path: str, console: Console):
    git = get_git_binary()
    try:
        repo = git.Repo(path=project_root_path, search_parent_directories=True)
        if repo.is_dirty():
            console.print(
                "[bold red]Error:[/] The repository has uncommitted changes. "
                "Please commit or stash them before proceeding."
            )
            sys.exit(1)
        current_active_branch = repo.active_branch.name
        console.print(
            f"[bold magenta]TrueFoundry:[/] Current branch [green]{current_active_branch!r}[/]"
        )
        branch_name = Prompt.ask(
            "[bold magenta]TrueFoundry:[/] Enter a branch name if you want to checkout to a new branch. "
            f"Press enter to continue on [green]{current_active_branch!r}[/]",
            console=console,
        )
        if branch_name:
            repo.git.checkout("-b", branch_name)
            console.print(
                f"[bold magenta]TrueFoundry:[/] Switched to branch: [green]{repo.active_branch}[/]"
            )
        else:
            console.print(
                f"[bold magenta]TrueFoundry:[/] Continuing on [green]{current_active_branch!r}[/]"
            )

    except git.exc.InvalidGitRepositoryError:
        console.print(
            "[red]Error:[/] This operation can only be performed inside a Git repository.\n"
            "Execute 'git init' to create a new repository."
        )
        sys.exit(1)

    except git.GitCommandError as gce:
        console.print(
            f"Command execution failed due to the following error:[red]{gce.stderr}[/]".replace(
                "\n  stderr:", ""
            )
        )
        console.print(
            "[bold red]Error:[/] Unable to switch to the new branch. It's possible that this branch already exists."
        )
        sys.exit(1)


def _update_status(event, status: Status, component_type: ComponentType):
    if isinstance(event, (AskQuestion, CommitConfirmation)):
        status.stop()

    if isinstance(
        event, (Developer.Request, ProjectIdentifier.Response, Tester.Response)
    ):
        status.update(
            "[bold magenta]TrueFoundry[/] is currently building the project. Please wait..."
        )

    if isinstance(event, ProjectIdentifier.Request):
        status.update(
            "[bold magenta]TrueFoundry[/] is currently identifying the project..."
        )

    if isinstance(event, (Tester.Request, DockerRun.Response)):
        status.update(
            "[bold magenta]TrueFoundry[/] is currently running tests on the project..."
        )

    if isinstance(event, DockerRunLog):
        if component_type == ComponentType.SERVICE:
            status.update(
                "[bold cyan]Running:[/] [bold magenta]TrueFoundry[/] is running your app in a Docker container. "
                "Press ctrl+c once your app is ready for testing."
            )
        else:
            status.update(
                "[bold cyan]Running:[/] [bold magenta]TrueFoundry[/] is running your app in a Docker container "
                "and waiting for completion."
            )


def _get_default_project_name(project_root_path: str):
    path = os.path.abspath(project_root_path).rstrip(os.path.sep)
    name = path.split(os.path.sep)[-1].lower()
    name = re.sub(r"[^a-z0-9]", "-", name)
    name = "-".join(n for n in name.split("-") if n)[:30]
    return name


def _get_docker(console: Console) -> docker.DockerClient:
    try:
        return docker.from_env()
    except Exception as e:
        raise InvalidRequirementsException(
            message="Could not connect to Docker, please check whether the Docker daemon is running."
        ) from e


def cli(project_root_path: str, deploy: bool, workspace_fqn: str = None):
    console = Console()
    openai_client = _get_openai_client()
    docker_client = _get_docker(console)
    project_root_path = os.path.abspath(project_root_path)
    console.print(ABOUT_AUTODEPLOY)
    console.print(
        "[bold reverse]You will need to have Docker and Git installed on your machine for this to work[/]"
    )
    if AUTODEPLOY_OPENAI_BASE_URL is not None and AUTODEPLOY_OPENAI_API_KEY is not None:
        console.print(
            "[bold green]OpenAI credentials found in environment variables.[/]"
        )
        console.print(
            "This operation will use tokens from your provided OpenAI account and may incur costs.",
        )
    else:
        console.print(
            "[dim]To use your own LLM, "
            "set the environment variables [dim italic green]AUTODEPLOY_OPENAI_BASE_URL[/],[/]",
            "[dim][dim italic green]AUTODEPLOY_OPENAI_API_KEY[/], "
            "and [dim italic green]AUTODEPLOY_MODEL_NAME[/] for URL, API key, and LLM model name respectively.[/]",
        )
    console.print(
        "[bold cyan]Note:[/] All changes will be committed to a new branch. Please ensure you have a repository."
    )
    console.print("[bright_green]Let's get started[/]")
    _check_repo(project_root_path=project_root_path, console=console)

    choices = {
        "Service: An application that runs continuously. "
        "Example: web servers, workers polling a job queue, etc.": "SERVICE",
        "Job: An application that runs once and then stops. "
        "Example: Training an ML model, running a script, etc.": "JOB",
    }
    component = questionary.select(
        "TrueFoundry: Is your project a", choices=list(choices.keys())
    ).ask()
    component_type = ComponentType[choices[component]]
    while True:
        name = Prompt.ask(
            "[bold magenta]TrueFoundry:[/] Name of deployment, or press [green]Enter[/] to select default. "
            f"default: [green]{_get_default_project_name(project_root_path)}",
            console=console,
            default=_get_default_project_name(project_root_path),
            show_default=False,
        )
        if not re.match(r"^[a-z][a-z0-9\-]{1,30}[a-z0-9]$", name):
            console.print(
                "[bold magenta]TrueFoundry:[/] The name should be between 2-30 alphanumeric"
                " characters and '-'. The first character should not be a digit."
            )
        else:
            break
    command = Prompt.ask(
        "[bold magenta]TrueFoundry:[/] Command to run the application or press [green]Enter[/] to skip",
        console=console,
        show_default=False,
        default=None,
    )

    env_path = Prompt.ask(
        "[bold magenta]TrueFoundry:[/] Enter .env file location for environment variables, "
        "or press [green]Enter[/] to skip.",
        console=console,
    )
    if workspace_fqn is None:
        workspace_fqn = Prompt.ask(
            "[bold magenta]TrueFoundry:[/] Enter the Workspace FQN where you would like to deploy, [dim]"
            "Ex: cluster-name:workspace-name[/]"
        )
    env = {}
    while True:
        try:
            env = _parse_env(project_root_path, env_path) if env_path else {}
            break
        except FileNotFoundError:
            console.print("[red]Invalid location for .env[/]")
            env_path = Prompt.ask(
                "[bold magenta]TrueFoundry:[/]Please provide the correct path,"
                "or press [green]Enter[/] to skip.",
                console=console,
            )
            continue
    status = console.status(
        "[bold cyan]Starting up:[/] [bold magenta]TrueFoundry[/] is initializing. Please wait..."
    )
    with status:
        developer = Developer(
            project_root_path=project_root_path,
            openai_client=openai_client,
            docker_client=docker_client,
            environment=env,
        )
        developer_run = developer.run(developer.Request(command=command, name=name))
        inp = None
        response = None
        while True:
            try:
                status.start()
                event = developer_run.send(inp)
                _update_status(
                    event=event, status=status, component_type=component_type
                )
                inp = event.render(console)
            except StopIteration as ex:
                response = ex.value
                break

    if deploy:
        console.rule("[bold green]Deploying to TrueFoundry[/]")
        deploy_component(
            workspace_fqn=workspace_fqn,
            project_root_path=project_root_path,
            dockerfile_path=response.dockerfile_path,
            name=name,
            component_type=component_type,
            env=env,
            command=response.command,
            port=response.port,
        )


@click.command(name="auto-deploy", cls=COMMAND_CLS)
@click.option(
    "--path", type=click.STRING, required=True, help="The root path of the project"
)
@click.option(
    "--deploy",
    type=click.BOOL,
    is_flag=True,
    default=True,
    show_default=True,
    help="Deploy the project after successfully building it.",
)
def autodeploy_cli(path: str, deploy: bool):
    """
    Build and deploy projects using TrueFoundry
    """
    cli(
        project_root_path=path,
        deploy=deploy,
    )



import time
from abc import ABC, abstractmethod
from typing import Optional

import requests

from truefoundry.common.constants import VERSION_PREFIX
from truefoundry.common.entities import DeviceCode, Token
from truefoundry.common.exceptions import BadRequestException
from truefoundry.common.request_utils import (
    request_handling,
    requests_retry_session,
)
from truefoundry.common.utils import poll_for_function, relogin_error_message
from truefoundry.logger import logger


class AuthServiceClient(ABC):
    def __init__(self, tenant_name: str):
        self._tenant_name = tenant_name

    @classmethod
    def from_tfy_host(cls, tfy_host: str) -> "AuthServiceClient":
        from truefoundry.common.servicefoundry_client import (
            ServiceFoundryServiceClient,
        )

        client = ServiceFoundryServiceClient(tfy_host=tfy_host)
        if client.python_sdk_config.use_sfy_server_auth_apis:
            return ServiceFoundryServerAuthServiceClient(
                tenant_name=client.tenant_info.tenant_name, url=client._api_server_url
            )
        return AuthServerServiceClient(
            tenant_name=client.tenant_info.tenant_name,
            url=client.tenant_info.auth_server_url,
        )

    @abstractmethod
    def refresh_token(self, token: Token, host: Optional[str] = None) -> Token: ...

    @abstractmethod
    def get_device_code(self) -> DeviceCode: ...

    @abstractmethod
    def get_token_from_device_code(
        self, device_code: str, timeout: float = 60, poll_interval_seconds: int = 1
    ) -> Token: ...


class ServiceFoundryServerAuthServiceClient(AuthServiceClient):
    def __init__(self, tenant_name: str, url: str):
        super().__init__(tenant_name=tenant_name)
        self._api_server_url = url

    def refresh_token(self, token: Token, host: Optional[str] = None) -> Token:
        host_arg_str = host if host else "HOST"
        if not token.refresh_token:
            # TODO: Add a way to propagate error messages without traceback to the output interface side
            raise Exception(
                relogin_error_message(
                    "Unable to resume login session.", host=host_arg_str
                )
            )
        url = f"{self._api_server_url}/{VERSION_PREFIX}/oauth2/token"
        data = {
            "tenantName": token.tenant_name,
            "refreshToken": token.refresh_token,
            "grantType": "refresh_token",
            "returnJWT": True,
        }
        session = requests_retry_session(retries=2)
        response = session.post(url, json=data)
        try:
            response_data = request_handling(response)
            return Token.parse_obj(response_data)
        except BadRequestException as ex:
            raise Exception(
                relogin_error_message(
                    "Unable to resume login session.", host=host_arg_str
                )
            ) from ex

    def get_device_code(self) -> DeviceCode:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/oauth2/device-authorize"
        data = {"tenantName": self._tenant_name}
        session = requests_retry_session(retries=2)
        response = session.post(url, json=data)
        response_data = request_handling(response)
        return DeviceCode.parse_obj(response_data)

    def get_token_from_device_code(
        self, device_code: str, timeout: float = 60, poll_interval_seconds: int = 1
    ) -> Token:
        timeout = timeout or 60
        poll_interval_seconds = poll_interval_seconds or 1
        url = f"{self._api_server_url}/{VERSION_PREFIX}/oauth2/token"
        data = {
            "tenantName": self._tenant_name,
            "deviceCode": device_code,
            "grantType": "device_code",
            "returnJWT": True,
        }
        start_time = time.monotonic()

        for response in poll_for_function(
            requests.post, poll_after_secs=poll_interval_seconds, url=url, json=data
        ):
            if response.status_code == 201:
                response = response.json()
                return Token.parse_obj(response)
            elif response.status_code == 202:
                logger.debug("User has not authorized yet. Checking again.")
            else:
                raise Exception(
                    "Failed to get token using device code. "
                    f"status_code {response.status_code},\n {response.text}"
                )
            time_elapsed = time.monotonic() - start_time
            if time_elapsed > timeout:
                logger.warning("Polled server for %s secs.", int(time_elapsed))
                break

        raise Exception(f"Did not get authorized within {timeout} seconds.")


class AuthServerServiceClient(AuthServiceClient):
    def __init__(self, tenant_name: str, url: str):
        super().__init__(tenant_name=tenant_name)

        self._auth_server_url = url

    def refresh_token(self, token: Token, host: Optional[str] = None) -> Token:
        host_arg_str = host if host else "HOST"
        if not token.refresh_token:
            # TODO: Add a way to propagate error messages without traceback to the output interface side
            raise Exception(
                relogin_error_message(
                    "Unable to resume login session.", host=host_arg_str
                )
            )
        url = f"{self._auth_server_url}/api/{VERSION_PREFIX}/oauth/token/refresh"
        data = {
            "tenantName": token.tenant_name,
            "refreshToken": token.refresh_token,
        }
        session = requests_retry_session(retries=2)
        response = session.post(url, json=data)
        try:
            response_data = request_handling(response)
            return Token.parse_obj(response_data)
        except BadRequestException as ex:
            raise Exception(
                relogin_error_message(
                    "Unable to resume login session.", host=host_arg_str
                )
            ) from ex

    def get_device_code(self) -> DeviceCode:
        url = f"{self._auth_server_url}/api/{VERSION_PREFIX}/oauth/device"
        data = {"tenantName": self._tenant_name}
        session = requests_retry_session(retries=2)
        response = session.post(url, json=data)
        response_data = request_handling(response)
        assert isinstance(response_data, dict)
        # TODO: temporary cleanup of incorrect attributes
        return DeviceCode.parse_obj(
            {
                "userCode": response_data.get("userCode"),
                "deviceCode": response_data.get("deviceCode"),
            }
        )

    def get_token_from_device_code(
        self, device_code: str, timeout: float = 60, poll_interval_seconds: int = 1
    ) -> Token:
        timeout = timeout or 60
        poll_interval_seconds = poll_interval_seconds or 1
        url = f"{self._auth_server_url}/api/{VERSION_PREFIX}/oauth/device/token"
        data = {
            "tenantName": self._tenant_name,
            "deviceCode": device_code,
        }
        start_time = time.monotonic()

        for response in poll_for_function(
            requests.post, poll_after_secs=poll_interval_seconds, url=url, json=data
        ):
            if response.status_code == 201:
                response = response.json()
                return Token.parse_obj(response)
            elif response.status_code == 202:
                logger.debug("User has not authorized yet. Checking again.")
            else:
                raise Exception(
                    "Failed to get token using device code. "
                    f"status_code {response.status_code},\n {response.text}"
                )
            time_elapsed = time.monotonic() - start_time
            if time_elapsed > timeout:
                logger.warning("Polled server for %s secs.", int(time_elapsed))
                break

        raise Exception(f"Did not get authorized within {timeout} seconds.")



import enum
import os
from pathlib import Path
from typing import Any, Dict, Optional

from truefoundry.pydantic_v1 import BaseSettings, Field, SecretStr

TFY_CONFIG_DIR = Path.home() / ".truefoundry"
CREDENTIAL_FILEPATH = TFY_CONFIG_DIR / "credentials.json"

# These keys are kept separately because we use them in error messages and some checks
TFY_HOST_ENV_KEY = "TFY_HOST"
TFY_API_KEY_ENV_KEY = "TFY_API_KEY"
TFY_DEBUG_ENV_KEY = "TFY_DEBUG"
TFY_INTERNAL_ENV_KEY = "TFY_INTERNAL"

TFY_INTERNAL_SIGNED_URL_SERVER_HOST_ENV_KEY = "TFY_INTERNAL_SIGNED_URL_SERVER_HOST"
TFY_INTERNAL_SIGNED_URL_SERVER_TOKEN_ENV_KEY = "TFY_INTERNAL_SIGNED_URL_SERVER_TOKEN"

# === Global Constants for OpenAI Integration ===
OPENAI_API_KEY_KEY = "OPENAI_API_KEY"
OPENAI_MODEL_KEY = "OPENAI_MODEL"

# === Ask Command Specific Environment Variables for TrueFoundry ===
TFY_ASK_OPENAI_API_KEY_KEY = "TFY_ASK_OPENAI_API_KEY"
TFY_ASK_OPENAI_BASE_URL_KEY = "TFY_ASK_OPENAI_BASE_URL"
TFY_ASK_MODEL_NAME_KEY = "TFY_ASK_OPENAI_MODEL"


class PythonPackageManager(str, enum.Enum):
    PIP = "pip"
    UV = "uv"


class TrueFoundrySdkEnv(BaseSettings):
    # Note: Every field in this class should have a default value
    # Never expect the user to set these values

    # For local development, this enables further configuration via _TFYServersConfig
    TFY_CLI_LOCAL_DEV_MODE: bool = False

    # These two are not meant to be used directly. See the TFY_HOST and TFY_API_KEY properties below
    TFY_HOST_: Optional[str] = Field(default=None, env=TFY_HOST_ENV_KEY)
    TFY_API_KEY_: Optional[SecretStr] = Field(default=None, env=TFY_API_KEY_ENV_KEY)

    ##############################
    # Internal Signed URL Server #
    ##############################
    TFY_INTERNAL_SIGNED_URL_SERVER_HOST: Optional[str] = Field(
        default=None, env=TFY_INTERNAL_SIGNED_URL_SERVER_HOST_ENV_KEY
    )
    TFY_INTERNAL_SIGNED_URL_SERVER_TOKEN: Optional[str] = Field(
        default=None, env=TFY_INTERNAL_SIGNED_URL_SERVER_TOKEN_ENV_KEY
    )
    TFY_INTERNAL_SIGNED_URL_SERVER_MAX_TIMEOUT: int = 5  # default: 5 seconds
    TFY_INTERNAL_SIGNED_URL_SERVER_DEFAULT_TTL: int = 3600  # default: 1 hour
    TFY_INTERNAL_MULTIPART_UPLOAD_FINALIZE_SIGNED_URL_TIMEOUT: int = (
        24 * 60 * 60
    )  # default: 24 hour
    TFY_INTERNAL_SIGNED_URL_REQUEST_TIMEOUT: int = 3600  # default: 1 hour
    TFY_INTERNAL_SIGNED_URL_CLIENT_LOG_LEVEL: str = "WARNING"

    #############
    # Artifacts #
    #############
    TFY_ARTIFACTS_DOWNLOAD_CHUNK_SIZE_BYTES: int = 100 * 1000 * 1000
    TFY_ARTIFACTS_DOWNLOAD_MAX_WORKERS: int = max(min(32, (os.cpu_count() or 2) * 2), 4)
    TFY_ARTIFACTS_UPLOAD_MAX_WORKERS: int = max(min(32, (os.cpu_count() or 2) * 2), 4)
    TFY_ARTIFACTS_DISABLE_MULTIPART_UPLOAD: bool = False
    TFY_ARTIFACTS_DOWNLOAD_FSYNC_CHUNKS: bool = False

    #############
    # TFY Build #
    #############
    # For customizing the images and packages used for builds
    TFY_PYTHONBUILD_PYTHON_IMAGE_REPO: str = "public.ecr.aws/docker/library/python"
    TFY_PYTHON_BUILD_PACKAGE_MANAGER: PythonPackageManager = PythonPackageManager.UV
    TFY_PYTHON_BUILD_UV_IMAGE_REPO: str = "ghcr.io/astral-sh/uv"
    TFY_PYTHON_BUILD_UV_IMAGE_TAG: str = "latest"
    TFY_PYTHON_BUILD_POETRY_VERSION: str = Field(
        default="2.0", env="TFY_PYTHON_BUILD_POETRY_VERSION"
    )
    TFY_PYTHON_BUILD_LATEST_POETRY_MAJOR_VERSION: int = Field(
        default=2, env="TFY_PYTHON_BUILD_LATEST_POETRY_MAJOR_VERSION"
    )
    TFY_SPARK_BUILD_SPARK_IMAGE_REPO: str = "public.ecr.aws/bitnami/spark"
    TFY_TASK_PYSPARK_BUILD_SPARK_IMAGE_REPO: str = "public.ecr.aws/bitnami/spark"

    ##############
    # OpenAI API #
    ##############
    # Global Constants for OpenAI Integration
    OPENAI_API_KEY: Optional[str] = Field(default=None, env=OPENAI_API_KEY_KEY)
    OPENAI_MODEL: Optional[str] = Field(default=None, env=OPENAI_MODEL_KEY)

    ###########
    # TFY Ask #
    ###########
    # Ask Command Specific Environment Variables for TrueFoundry
    TFY_ASK_OPENAI_API_KEY: Optional[str] = Field(
        default=None, env=TFY_ASK_OPENAI_API_KEY_KEY
    )
    TFY_ASK_OPENAI_BASE_URL: Optional[str] = Field(
        default=None, env=TFY_ASK_OPENAI_BASE_URL_KEY
    )
    TFY_ASK_MODEL_NAME: Optional[str] = Field(default=None, env=TFY_ASK_MODEL_NAME_KEY)
    TFY_ASK_GENERATION_PARAMS: Dict[str, Any] = Field(
        default_factory=lambda: {"temperature": 0.0, "top_p": 1, "max_tokens": 4096}
    )
    TFY_ASK_SYSTEM_PROMPT_NAME: str = Field(default="tfy-ask-k8s-prompt")
    TFY_INTERNAL_ASK_CONFIG_OVERRIDE_FILE: Optional[str] = Field(default=None)

    # This is a hack to fresh read the env vars because people can end up importing this file
    # before setting the correct env vars. E.g. in notebook environments.
    @property
    def TFY_HOST(self) -> Optional[str]:
        self.__init__()
        return self.TFY_HOST_

    @property
    def TFY_API_KEY(self) -> Optional[SecretStr]:
        self.__init__()
        return self.TFY_API_KEY_


ENV_VARS = TrueFoundrySdkEnv()
API_SERVER_RELATIVE_PATH = "api/svc"
MLFOUNDRY_SERVER_RELATIVE_PATH = "api/ml"
VERSION_PREFIX = "v1"
SERVICEFOUNDRY_CLIENT_MAX_RETRIES = 2



"""
Utilities for validating user inputs such as metric names and parameter names.
"""

import json
import numbers
import posixpath
import re
from operator import xor
from typing import List, Optional, Union

from truefoundry.ml import MlFoundryException
from truefoundry.ml._autogen.client import (  # type: ignore[attr-defined]
    MetricDto,
    ParamDto,
    RunTagDto,
)

_VALID_PARAM_AND_METRIC_NAMES = re.compile(r"^[/\w.\- ]*$")

# Regex for valid run IDs: must be an alphanumeric string of length 1 to 256.
_BAD_CHARACTERS_MESSAGE = (
    "Names may only contain alphanumerics, underscores (_), dashes (-), periods (.),"
    " spaces ( ), and slashes (/)."
)
_RUN_ID_REGEX = re.compile(r"^[a-zA-Z0-9][\w\-]{0,255}$")
_ML_REPO_ID_REGEX = re.compile(r"^[a-zA-Z0-9][\w\-]{0,63}$")
_ML_REPO_NAME_REGEX = re.compile(r"^[a-zA-Z][a-zA-Z0-9\-]{1,98}[a-zA-Z0-9]$")
_RUN_NAME_REGEX = re.compile(r"^[a-zA-Z0-9-]*$")
_RUN_LOG_LOG_TYPE_REGEX = re.compile(r"^[a-zA-Z0-9-/]*$")
_RUN_LOG_KEY_REGEX = re.compile(r"^[a-zA-Z0-9-_]*$")
_APP_NAME_REGEX = re.compile(r"^[a-z][a-z0-9\\-]{1,30}[a-z0-9]$")


MAX_PARAMS_TAGS_PER_BATCH = 100
MAX_METRICS_PER_BATCH = 1000
MAX_ENTITIES_PER_BATCH = 1000
MAX_BATCH_LOG_REQUEST_SIZE = int(1e6)
MAX_PARAM_VAL_LENGTH = 1000
MAX_TAG_VAL_LENGTH = 5000
MAX_ML_REPO_TAG_KEY_LENGTH = 250
MAX_ML_REPO_TAG_VAL_LENGTH = 5000
MAX_ENTITY_KEY_LENGTH = 250
MAX_RUN_LOG_KEY_NAME_LENGTH = 128
MAX_RUN_LOG_LOG_TYPE_LENGTH = 128
MAX_ML_REPO_DESCRIPTION_LENGTH = 512
MAX_RUN_DESCRIPTION_LENGTH = 512
MAX_ML_REPOS_LISTED_PER_PAGE = 100


def is_string_type(item):
    return isinstance(item, str)


def bad_path_message(name: str):
    return (
        f"Names may be treated as files in certain cases, "
        f"and must not resolve to other names when treated as such. "
        f"This name would resolve to {posixpath.normpath(name)!r}"
    )


def path_not_unique(name: str):
    norm = posixpath.normpath(name)
    return norm != name or norm == "." or norm.startswith("..") or norm.startswith("/")


def _validate_metric_name(name: str):
    """Check that `name` is a valid metric name and raise an exception if it isn't."""
    if name is None or not _VALID_PARAM_AND_METRIC_NAMES.match(name):
        raise MlFoundryException(
            f"Invalid metric name: {name!r}. {_BAD_CHARACTERS_MESSAGE}",
        )
    if path_not_unique(name):
        raise MlFoundryException(
            f"Invalid metric name: {name!r}. {bad_path_message(name)}",
        )


def _validate_run_name(name: Optional[str]):
    if not name or not _RUN_NAME_REGEX.match(name):
        raise MlFoundryException(
            f"Invalid run name {name!r}. Name should not empty string "
            f"and may only contain alphanumerics or dashes (-)",
        )
    _validate_length_limit("Run name", MAX_ENTITY_KEY_LENGTH, name)


def _is_numeric(value):
    """
    Returns True if the passed-in value is numeric.
    """
    # Note that `isinstance(bool_value, numbers.Number)` returns `True` because `bool` is a
    # subclass of `int`.
    return not isinstance(value, bool) and isinstance(value, numbers.Number)


def _validate_metric(
    key: str,
    value: Optional[Union[int, float]],
    timestamp: Optional[int],
    step: Optional[int],
):
    """
    Check that a param with the specified key, value, timestamp is valid and raise an exception if
    it isn't.
    """
    _validate_metric_name(key)
    _validate_length_limit("Metric name", MAX_ENTITY_KEY_LENGTH, key)
    # value must be a Number
    # since bool is an instance of Number check for bool additionally
    if not _is_numeric(value):
        raise MlFoundryException(
            f"Got invalid value {value} for metric {key!r} (step={step}). "
            f"Please specify value as a valid double (64-bit floating point)",
        )

    if not isinstance(timestamp, int) or timestamp < 0:
        raise MlFoundryException(
            f"Got invalid timestamp {timestamp} for metric {key!r} (value={value}). "
            f"Timestamp must be a non-negative long (64-bit integer)",
        )

    if not isinstance(step, int):
        raise MlFoundryException(
            f"Got invalid step {step} for metric {key!r} (value={value}). "
            f"Step must be a valid long (64-bit integer).",
        )


def _validate_run_log_input(
    key: str,
    timestamp: int,
    step: int,
    log_type: str,
    value: Optional[str],
    artifact_path: Optional[str],
):
    # reusing the validation defined for metric: str
    _validate_metric(key=key, value=0, timestamp=timestamp, step=step)
    _validate_length_limit("RunLog log type", MAX_RUN_LOG_LOG_TYPE_LENGTH, log_type)
    _validate_length_limit("RunLog log key", MAX_RUN_LOG_KEY_NAME_LENGTH, key)
    if not log_type or not _RUN_LOG_LOG_TYPE_REGEX.match(log_type):
        raise MlFoundryException(
            f"Invalid run log_type: {log_type}",
        )

    if not key or not _RUN_LOG_KEY_REGEX.match(key):
        raise MlFoundryException(
            f"Invalid run log key: {key!r} should only contain alphanumeric, hyphen or underscore"
        )

    if not xor(bool(artifact_path), bool(value)):
        raise MlFoundryException(
            "Either artifact_path or value should be empty. "
            f"artifact_path={artifact_path!r}, value={value!r}",
        )
    if value is not None:
        try:
            json.loads(value)
        except ValueError as e:
            raise MlFoundryException(
                f"Key {key!r} does not contain a valid json: {e}",
            ) from e


def _validate_param(key: str, value: str):
    """
    Check that a param with the specified key & value is valid and raise an exception if it
    isn't.
    """
    _validate_param_name(key)
    _validate_length_limit("Param key", MAX_ENTITY_KEY_LENGTH, key)
    _validate_length_limit("Param value", MAX_PARAM_VAL_LENGTH, value)


def _validate_tag(key: str, value: str):
    """
    Check that a tag with the specified key & value is valid and raise an exception if it isn't.
    """
    _validate_tag_name(key)
    _validate_length_limit("Tag key", MAX_ENTITY_KEY_LENGTH, key)
    _validate_length_limit("Tag value", MAX_TAG_VAL_LENGTH, value)


def _validate_ml_repo_tag(key: str, value: str):
    """
    Check that a tag with the specified key & value is valid and raise an exception if it isn't.
    """
    _validate_tag_name(key)
    _validate_length_limit("Tag key", MAX_ML_REPO_TAG_KEY_LENGTH, key)
    _validate_length_limit("Tag value", MAX_ML_REPO_TAG_VAL_LENGTH, value)


def _validate_list_ml_repos_max_results(max_results):
    """
    Check that `max_results` is within an acceptable range and raise an exception if it isn't.
    """
    if max_results is None:
        return

    if max_results < 1:
        raise MlFoundryException(
            f"Invalid value for request parameter max_results. "
            f"It must be at least 1, but got value {max_results}",
        )

    if max_results > MAX_ML_REPOS_LISTED_PER_PAGE:
        raise MlFoundryException(
            f"Invalid value for request parameter max_results. "
            f"It must be at most {MAX_ML_REPOS_LISTED_PER_PAGE}, but got value {max_results}",
        )


def _validate_param_keys_unique(params: List[ParamDto]):
    """Ensures that duplicate param keys are not present in the `log_batch()` params argument"""
    unique_keys = []
    dupe_keys = []
    for param in params:
        if param.key not in unique_keys:
            unique_keys.append(param.key)
        else:
            dupe_keys.append(param.key)

    if dupe_keys:
        raise MlFoundryException(
            f"Duplicate parameter keys have been submitted: {dupe_keys}. Please ensure "
            f"the request contains only one param value per param key.",
        )


def _validate_param_name(name: str):
    """Check that `name` is a valid parameter name and raise an exception if it isn't."""
    if name is None or not _VALID_PARAM_AND_METRIC_NAMES.match(name):
        raise MlFoundryException(
            f"Invalid parameter name: {name!r}. {_BAD_CHARACTERS_MESSAGE}",
        )
    if path_not_unique(name):
        raise MlFoundryException(
            f"Invalid parameter name: {name!r}. {bad_path_message(name)}",
        )


def _validate_tag_name(name: str):
    """Check that `name` is a valid tag name and raise an exception if it isn't."""
    # Reuse param & metric check.
    if name is None or not _VALID_PARAM_AND_METRIC_NAMES.match(name):
        raise MlFoundryException(
            f"Invalid tag name: {name!r}. {_BAD_CHARACTERS_MESSAGE}",
        )
    if path_not_unique(name):
        raise MlFoundryException(
            f"Invalid tag name: {name!r}. {bad_path_message(name)}",
        )


def _validate_length_limit(entity_name: str, limit: int, value: str):
    if len(value) > limit:
        raise MlFoundryException(
            f"{entity_name} {value!r} had length {len(value)} which exceeded length limit of {limit}",
        )


def _validate_run_id(run_id: str):
    """Check that `run_id` is a valid run ID and raise an exception if it isn't."""
    if _RUN_ID_REGEX.match(run_id) is None:
        raise MlFoundryException(
            f"Invalid run ID: {run_id!r}",
        )


def _validate_ml_repo_id(ml_repo_id: str):
    """Check that `ml_repo_id`is a valid string or None, raise an exception if it isn't."""
    if ml_repo_id is not None and _ML_REPO_ID_REGEX.match(ml_repo_id) is None:
        raise MlFoundryException(
            f"Invalid ml_repo ID: {ml_repo_id!r}",
        )


def _validate_batch_limit(entity_name: str, limit: int, length: int):
    if length > limit:
        error_msg = (
            f"A batch logging request can contain at most {limit} {entity_name}. "
            f"Got {length} {entity_name}. "
            f"Please split up {entity_name} across multiple requests and try again."
        )
        raise MlFoundryException(error_msg)


def _validate_batch_log_limits(
    metrics: List[MetricDto], params: List[ParamDto], tags: List[RunTagDto]
):
    """Validate that the provided batched logging arguments are within expected limits."""
    _validate_batch_limit(
        entity_name="metrics", limit=MAX_METRICS_PER_BATCH, length=len(metrics)
    )
    _validate_batch_limit(
        entity_name="params", limit=MAX_PARAMS_TAGS_PER_BATCH, length=len(params)
    )
    _validate_batch_limit(
        entity_name="tags", limit=MAX_PARAMS_TAGS_PER_BATCH, length=len(tags)
    )
    total_length = len(metrics) + len(params) + len(tags)
    _validate_batch_limit(
        entity_name="metrics, params, and tags",
        limit=MAX_ENTITIES_PER_BATCH,
        length=total_length,
    )


def _validate_batch_log_data(
    metrics: List[MetricDto], params: List[ParamDto], tags: List[RunTagDto]
):
    if len(metrics) == 0 and len(params) == 0 and len(tags) == 0:
        return
    _validate_batch_log_limits(metrics=metrics, params=params, tags=tags)
    for metric in metrics:
        _validate_metric(metric.key, metric.value, metric.timestamp, metric.step)
    if len(params) > 1:
        _validate_param_keys_unique(params)
    for param in params:
        _validate_param(param.key, param.value)
    for tag in tags:
        _validate_tag(tag.key, tag.value)


def _validate_ml_repo_name(ml_repo_name: str):
    """Check that `ml_repo_name` is a valid string and raise an exception if it isn't."""
    if ml_repo_name == "" or ml_repo_name is None or not is_string_type(ml_repo_name):
        raise MlFoundryException(
            f"ml_repo must be string type and not empty. "
            f"Got {type(ml_repo_name)} type with value {ml_repo_name!r}"
        )

    if not _ML_REPO_NAME_REGEX.match(ml_repo_name):
        raise MlFoundryException(
            f"Invalid ML Repo name {ml_repo_name!r}. Name may only contain alphanumerics or dashes (-)",
        )


def _validate_ml_repo_description(description: str):
    _validate_length_limit(
        "ML Repo Description", MAX_ML_REPO_DESCRIPTION_LENGTH, description
    )


def _validate_run_description(description: str):
    _validate_length_limit("Run Description", MAX_RUN_DESCRIPTION_LENGTH, description)



# flake8: noqa

# import apis into api package
from truefoundry.ml._autogen.client.api.auth_api import AuthApi
from truefoundry.ml._autogen.client.api.deprecated_api import DeprecatedApi
from truefoundry.ml._autogen.client.api.experiments_api import ExperimentsApi
from truefoundry.ml._autogen.client.api.generate_code_snippet_api import (
    GenerateCodeSnippetApi,
)
from truefoundry.ml._autogen.client.api.health_api import HealthApi
from truefoundry.ml._autogen.client.api.metrics_api import MetricsApi
from truefoundry.ml._autogen.client.api.mlfoundry_artifacts_api import (
    MlfoundryArtifactsApi,
)
from truefoundry.ml._autogen.client.api.run_artifacts_api import RunArtifactsApi
from truefoundry.ml._autogen.client.api.runs_api import RunsApi



import warnings


class TrueFoundryDeprecationWarning(DeprecationWarning):
    pass


def surface_truefoundry_deprecation_warnings() -> None:
    """Unmute TrueFoundry deprecation warnings."""
    warnings.filterwarnings(
        "default",
        category=TrueFoundryDeprecationWarning,
    )


def suppress_truefoundry_deprecation_warnings() -> None:
    """Mute TrueFoundry deprecation warnings."""
    warnings.filterwarnings(
        "ignore",
        category=TrueFoundryDeprecationWarning,
    )


def suppress_truefoundry_ml_autogen_warnings() -> None:
    """Mute TrueFoundry ML autogen warnings."""
    warnings.filterwarnings(
        "ignore",
        module="truefoundry.ml._autogen",
    )



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, Dict, Optional

from truefoundry.ml._autogen.client.models.experiment_tag_dto import ExperimentTagDto
from truefoundry.ml._autogen.client.models.subject import Subject
from truefoundry.pydantic_v1 import BaseModel, Field, StrictInt, StrictStr, conlist


class ExperimentDto(BaseModel):
    """
    ExperimentDto
    """

    experiment_id: StrictStr = Field(...)
    name: StrictStr = Field(...)
    storage_integration_id: Optional[StrictStr] = None
    artifact_location: Optional[StrictStr] = None
    lifecycle_stage: Optional[StrictStr] = None
    tags: Optional[conlist(ExperimentTagDto)] = None
    creator_user_id: StrictStr = Field(...)
    created_by_subject: Subject = Field(...)
    description: Optional[StrictStr] = None
    privacy_type: Optional[StrictStr] = "PRIVATE"
    created_at: Optional[StrictInt] = None
    tenant_name: Optional[StrictStr] = None
    manifest: Dict[str, Any] = Field(...)
    num_runs: Optional[StrictInt] = None
    artifact_type_counts: Optional[Dict[str, StrictInt]] = None
    datasets_count: Optional[StrictInt] = None
    num_user_collaborators: Optional[StrictInt] = None
    __properties = [
        "experiment_id",
        "name",
        "storage_integration_id",
        "artifact_location",
        "lifecycle_stage",
        "tags",
        "creator_user_id",
        "created_by_subject",
        "description",
        "privacy_type",
        "created_at",
        "tenant_name",
        "manifest",
        "num_runs",
        "artifact_type_counts",
        "datasets_count",
        "num_user_collaborators",
    ]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ExperimentDto:
        """Create an instance of ExperimentDto from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of each item in tags (list)
        _items = []
        if self.tags:
            for _item in self.tags:
                if _item:
                    _items.append(_item.to_dict())
            _dict["tags"] = _items
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of created_by_subject
        if self.created_by_subject:
            _dict["created_by_subject"] = self.created_by_subject.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> ExperimentDto:
        """Create an instance of ExperimentDto from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return ExperimentDto.parse_obj(obj)

        _obj = ExperimentDto.parse_obj(
            {
                "experiment_id": obj.get("experiment_id"),
                "name": obj.get("name"),
                "storage_integration_id": obj.get("storage_integration_id"),
                "artifact_location": obj.get("artifact_location"),
                "lifecycle_stage": obj.get("lifecycle_stage"),
                "tags": [ExperimentTagDto.from_dict(_item) for _item in obj.get("tags")]
                if obj.get("tags") is not None
                else None,
                "creator_user_id": obj.get("creator_user_id"),
                "created_by_subject": Subject.from_dict(obj.get("created_by_subject"))
                if obj.get("created_by_subject") is not None
                else None,
                "description": obj.get("description"),
                "privacy_type": obj.get("privacy_type")
                if obj.get("privacy_type") is not None
                else "PRIVATE",
                "created_at": obj.get("created_at"),
                "tenant_name": obj.get("tenant_name"),
                "manifest": obj.get("manifest"),
                "num_runs": obj.get("num_runs"),
                "artifact_type_counts": obj.get("artifact_type_counts"),
                "datasets_count": obj.get("datasets_count"),
                "num_user_collaborators": obj.get("num_user_collaborators"),
            }
        )
        return _obj
